﻿using UnityEngine;
using UnityEngine.SceneManagement;
using System.Collections;
using System.Collections.Generic;

public class MovementManager : MonoBehaviour
{
    public static MovementManager Instance { get; private set; }

    public float DelayBeforeStart = 1f;
    public int MaximumPoints = 20;
    public GameObject Player;
    public int CurrentLevelIndex;
    public LevelConfig[] AllLevels;
    public float MusicBPM = 130f;
    public float LoopBeatCount = 32f;
    public float BackBeatCount = 4f;

    public LevelConfig CurrentLevel
    {
        get { return AllLevels[CurrentLevelIndex]; }	
    }

    private float timeBetweenMoves;
    private float lastMoveTime = float.MaxValue;
    private bool usingBackBeat1 = true;
    private float timeBetweenBackBeats;
    private float lastBackBeatTime = float.MaxValue;
    private bool usingLoop1 = true;
    private float timeBetweenLoops;
    private float lastLoopTime = float.MaxValue;
    private int retries;

    public delegate void DoNextAction(LevelConfig.Actions action, LevelConfig.CubiclePosition previousPosition, LevelConfig.CubiclePosition newPosition, float cubicleObjectHitDirection);
    public event DoNextAction OnDoNextAction;

    public int CurrentIndex { get; private set; }
    public LevelConfig.Actions CurrentAction { get; private set; }
    public LevelConfig.CubiclePosition PreviousPosition { get; private set; }
    public LevelConfig.CubiclePosition CurrentPosition { get; private set; }
    public float CurrentCubicleObjectHitDirection { get; private set; }

    void Awake()
    {
        Instance = this;

        NotificationServer.Instance.AddObserver(gameObject, NotificationName.Unpause, OnUnpause);

        for (int i = 0; i < AllLevels.Length; i++)
        {
            if (!(Mathf.Approximately(AllLevels[i].MoveSpeed, 1f) || Mathf.Approximately(AllLevels[i].MoveSpeed, 2f) || Mathf.Approximately(AllLevels[i].MoveSpeed, 4f)))
                Debug.LogError("Level named '" + AllLevels[i].name + "' has a Move Speed of '" + AllLevels[i].MoveSpeed + "' which is not an allowed value (1, 2, or 4)");

            if ((AllLevels[i].MoveSets.Count % 4) != 0)
                Debug.LogError("Level named '" + AllLevels[i].name + "' has a Move Sets count that isn't a multiple of 4");
            
            AllLevels[i].SetUpMoves();
        }

        StartCoroutine(DelayThenStart());
    }

    void OnDestroy()
    {
        NotificationServer.Instance.RemoveObserver(gameObject, NotificationName.Unpause);
    }

    void Update()
    {
        UpdateMusic();

        UpdateMoves();
    }

    private void OnUnpause()
    {
        lastMoveTime = float.MaxValue;
        lastBackBeatTime = float.MaxValue;

        StopAllCoroutines();

        StartCoroutine(DelayThenStart());
    }

    private IEnumerator DelayThenStart()
    {
        yield return new WaitForSeconds(DelayBeforeStart);

        OnStartLevel();
    }

    private void UpdateMusic()
    {
        if ((Time.time - lastBackBeatTime) >= timeBetweenBackBeats)
        {
            if (usingBackBeat1)
                SoundManager.Instance.PlayBackgroundBeat1();
            else
                SoundManager.Instance.PlayBackgroundBeat2();

            usingBackBeat1 = !usingBackBeat1;

            if (Mathf.Approximately(lastBackBeatTime, float.MinValue))
                lastBackBeatTime = Time.time;
            else
                lastBackBeatTime += timeBetweenBackBeats;
        }

        if ((Time.time - lastLoopTime) >= timeBetweenBackBeats)
        {
            if (usingLoop1)
                SoundManager.Instance.PlayLoop1(CurrentLevel.MusicLoop);
            else
                SoundManager.Instance.PlayLoop2(CurrentLevel.MusicLoop);

            usingLoop1 = !usingLoop1;

            if (Mathf.Approximately(lastLoopTime, float.MinValue))
                lastLoopTime = Time.time;
            else
                lastLoopTime += timeBetweenLoops;
        }
    }

    private void UpdateMoves()
    {
        if ((Time.time - lastMoveTime) >= timeBetweenMoves)
        {
            CurrentAction = CurrentLevel.Moves[CurrentIndex];

            PreviousPosition = CurrentPosition;

            if (CurrentAction == LevelConfig.Actions.LeftPosition)
                CurrentPosition = LevelConfig.CubiclePosition.Left;
            if (CurrentAction == LevelConfig.Actions.CenterPosition)
                CurrentPosition = LevelConfig.CubiclePosition.Center;
            if (CurrentAction == LevelConfig.Actions.RightPosition)
                CurrentPosition = LevelConfig.CubiclePosition.Right;

            // Randomise the direction of the hit by randomly getting either -1 or 1
            CurrentCubicleObjectHitDirection = Random.Range(0, 2) == 0 ? -1 : 1;

            if (OnDoNextAction != null)
            {
                OnDoNextAction(CurrentAction, PreviousPosition, CurrentPosition, CurrentCubicleObjectHitDirection);
            }

            CurrentIndex++;
            if (CurrentIndex >= CurrentLevel.Moves.Count)
            {
                CurrentIndex = 0;

                if (Player != null)
                {
                    var p = Player.GetComponent<PlayerMover>();

                    if (p.Points >= CurrentLevel.PointsToPass && retries >= CurrentLevel.MinimumTries)
                    {
                        CurrentLevelIndex++;
                        if (CurrentLevelIndex >= AllLevels.Length)
                        {
                            CurrentLevelIndex = AllLevels.Length - 1;
                            // WOW YOU WIN!
                            FindObjectOfType<GameScene>().StartLoadGameOver();
                        }
                        else
                        {
                            OnStartLevel();
                            return;
                        }
                    }
                    else if (p.Points < CurrentLevel.PointsToRegress && retries >= CurrentLevel.MinimumTries)
                    {
                        CurrentLevelIndex--;
                        if (CurrentLevelIndex < 0)
                            CurrentLevelIndex = 0;
                        OnStartLevel();
                        return;
                    }
                    else
                        OnRestartLevel();
                }
            }

            if (Mathf.Approximately(lastMoveTime, float.MinValue))
                lastMoveTime = Time.time;
            else
                lastMoveTime += timeBetweenMoves;
        }
    }

    void OnRestartLevel()
    {
        retries++;
    }

    void OnStartLevel()
    {
        retries = 0;

        timeBetweenMoves = 60f / CurrentLevel.MovesPerMinute;
        lastMoveTime = Time.time - timeBetweenMoves * (1f / 64f);

        timeBetweenBackBeats = 60f / MusicBPM * BackBeatCount;
        lastBackBeatTime = Time.time - timeBetweenBackBeats * (3f / 4f);

        timeBetweenLoops = 60f / MusicBPM * LoopBeatCount;
        lastLoopTime = Time.time - timeBetweenBackBeats * (3f / 4f);

        CurrentIndex = 0;
        CurrentAction = LevelConfig.Actions.Nothing;
        PreviousPosition = LevelConfig.CubiclePosition.Center;
        CurrentPosition = LevelConfig.CubiclePosition.Center;

        if (Player != null)
        {
            var p = Player.GetComponent<PlayerMover>();
            p.TargetStage = CurrentLevel.EnvironmentStage;
        }
    }
}
