﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class Mover : MonoBehaviour
{
    public int Order;
    public Sprite NormalSprite;
	public Sprite UseCubicleObjectSprite;
	public Sprite MonkNormalSprite;
	public Sprite MonkUseCubicleObjectSprite;
    public Sprite FailSprite;
    public GameObject TutorialGameObject;
    public SpriteRenderer TutorialIcon;
    public GameObject CubicleObjectLeft;
    public GameObject CubicleObjectCenter;
    public GameObject CubicleObjectRight;
    public GameObject SmokePuff1;
    public GameObject SmokePuff2;

    protected Vector3 startPosition;
    protected LevelConfig.CubiclePosition currentPosition = LevelConfig.CubiclePosition.Center;
    protected LevelConfig.Actions currentAction = LevelConfig.Actions.Nothing;
    protected float actionStartTime;

    private Coroutine moveCoroutine;
    private Coroutine hitObjectCoroutine;
    private Coroutine tutorialCoroutine;
    private Coroutine smokePuffCoroutine1;
    private Coroutine smokePuffCoroutine2;
    private SpriteRenderer spriteRenderer;
    private SpriteSwapper spriteSwapper;
    private Vector3 startMovePosition;
    private Vector3 endMovePosition;
    private GameObject lastCubicleObject;
    private bool usingSmokePuff1 = true;
    private Vector3 startSmokePuffPosition;

    protected virtual void Start()
    {
        NotificationServer.Instance.AddObserver(gameObject, NotificationName.Unpause, OnUnpause);

        if (TutorialGameObject != null)
            TutorialGameObject.SetActive(false);

        startPosition = transform.localPosition;
        endMovePosition = startPosition;

        spriteSwapper = gameObject.GetComponent<SpriteSwapper>();
        spriteRenderer = GetComponent<SpriteRenderer>();

        startSmokePuffPosition = SmokePuff1.transform.localPosition;
        SmokePuff1.SetActive(false);
        SmokePuff2.SetActive(false);
    }

    protected virtual void OnDestroy()
    {
        NotificationServer.Instance.RemoveObserver(gameObject, NotificationName.Unpause);
    }

    protected virtual void OnUnpause()
    {
        StopAllCoroutines();

        ResetPosition();
        ResetCubicleObject();
        ResetTutorial();

        spriteRenderer.flipX = false;
        transform.localPosition = startPosition;
        startMovePosition = startPosition;
        endMovePosition = startPosition;

        currentPosition = LevelConfig.CubiclePosition.Center;
        currentAction = LevelConfig.Actions.Nothing;
    }

    protected void StartMovePosition(float direction)
    {
        if (moveCoroutine != null)
            StopCoroutine(moveCoroutine);
        
        moveCoroutine = StartCoroutine(MovePosition(direction));

        if (usingSmokePuff1 && smokePuffCoroutine2 != null)
            StopCoroutine(smokePuffCoroutine2);
        if (!usingSmokePuff1 && smokePuffCoroutine1 != null)
            StopCoroutine(smokePuffCoroutine1);

        if (usingSmokePuff1)
            smokePuffCoroutine2 = StartCoroutine(MoveSmokePuff(direction));
        else
            smokePuffCoroutine1 = StartCoroutine(MoveSmokePuff(direction));
    }

    protected IEnumerator MovePosition(float direction)
    {
        spriteRenderer.flipX = direction < 0f;

        float moveStartTime = Time.time;
        startMovePosition = new Vector3(transform.localPosition.x, transform.localPosition.y, transform.localPosition.z);
        endMovePosition = new Vector3(endMovePosition.x + GlobalSettings.Instance.MoveDistanceHorizontal * direction, endMovePosition.y, endMovePosition.z);

        while ((Time.time - moveStartTime) <= (GlobalSettings.Instance.MoveSpeed / 2f))
        {
            float zRotation = Mathf.SmoothStep(0f, GlobalSettings.Instance.MoveRotation * direction, (Time.time - moveStartTime) / (GlobalSettings.Instance.MoveSpeed / 2f));

            transform.localPosition = Vector3.Lerp(startMovePosition, endMovePosition, Mathf.SmoothStep(0f, 1f, (Time.time - moveStartTime) / GlobalSettings.Instance.MoveSpeed));
            transform.transform.localEulerAngles = new Vector3(0f, 0f, zRotation);

            yield return null;
        }

        while ((Time.time - moveStartTime) <= (GlobalSettings.Instance.MoveSpeed))
        {
            float zRotation = Mathf.SmoothStep(GlobalSettings.Instance.MoveRotation * direction, 0f, (Time.time - moveStartTime - GlobalSettings.Instance.MoveSpeed / 2f) / (GlobalSettings.Instance.MoveSpeed / 2f));

            transform.localPosition = Vector3.Lerp(startMovePosition, endMovePosition, Mathf.SmoothStep(0f, 1f, (Time.time - moveStartTime) / GlobalSettings.Instance.MoveSpeed));
            transform.transform.localEulerAngles = new Vector3(0f, 0f, zRotation);

            yield return null;
        }

        // Make sure to set the final value as the lerp might not get all the way there
        ResetPosition();

        yield return null;
    }

    protected IEnumerator MoveSmokePuff(float direction)
    {
        spriteRenderer.flipX = direction < 0f;

        float moveStartTime = Time.time;

        GameObject smokePuffToUse = usingSmokePuff1 ? SmokePuff1 : SmokePuff2;
        usingSmokePuff1 = !usingSmokePuff1;
        smokePuffToUse.SetActive(true);

        smokePuffToUse.transform.localPosition = new Vector3(startSmokePuffPosition.x * direction, startSmokePuffPosition.y, startSmokePuffPosition.z);

        Vector3 startScale = Vector3.zero;
        Vector3 endScale = Vector3.one;

        while ((Time.time - moveStartTime) <= GlobalSettings.Instance.SmokePuffGrowSpeed)
        {
            smokePuffToUse.transform.localScale = Vector3.Lerp(startScale, endScale, Mathf.SmoothStep(0f, 1f, (Time.time - moveStartTime) / GlobalSettings.Instance.SmokePuffGrowSpeed));

            yield return null;
        }

        moveStartTime = Time.time;

        while ((Time.time - moveStartTime) <= GlobalSettings.Instance.SmokePuffShrinkSpeed)
        {
            smokePuffToUse.transform.localScale = Vector3.Lerp(endScale, startScale, Mathf.SmoothStep(0f, 1f, (Time.time - moveStartTime) / GlobalSettings.Instance.SmokePuffShrinkSpeed));

            yield return null;
        }

        smokePuffToUse.SetActive(false);

        yield return null;
    }

    private void ResetPosition()
    {
        transform.localPosition = endMovePosition;
        transform.localEulerAngles = new Vector3(0f, 0f ,0f);
    }

    protected void StartHitCubicleObject(GameObject cubicleObject, float cubicleObjectHitDirection)
    {
        if (hitObjectCoroutine != null)
            StopCoroutine(hitObjectCoroutine);
        
        hitObjectCoroutine = StartCoroutine(HitCubicleObject(cubicleObject, cubicleObjectHitDirection));
    }

	protected bool IsMonkTime { 
		get {
			//return false;
			return MovementManager.Instance.CurrentLevel.EnvironmentStage >= 2;
		}
	}

    protected IEnumerator HitCubicleObject(GameObject cubicleObject, float cubicleObjectHitDirection)
    {
        lastCubicleObject = cubicleObject;

		spriteRenderer.sprite = IsMonkTime ? MonkUseCubicleObjectSprite : UseCubicleObjectSprite;

        float hitStartTime = Time.time;
        while ((Time.time - hitStartTime) <= GlobalSettings.Instance.CubicleObjectHitTime)
        {
            float zRotation = Mathf.SmoothStep(0f, GlobalSettings.Instance.CubicleObjectHitRotation, (Time.time - hitStartTime) / GlobalSettings.Instance.CubicleObjectHitTime);
            float xScale = Mathf.SmoothStep(1f, 1f - GlobalSettings.Instance.CubicleObjectHitSquish, (Time.time - hitStartTime) / GlobalSettings.Instance.CubicleObjectHitTime);
			cubicleObject.transform.localEulerAngles = new Vector3(0f, 0f, zRotation * cubicleObjectHitDirection);
			if (spriteSwapper == null || !spriteSwapper.IsAnimating)
				cubicleObject.transform.localScale = new Vector3(xScale, 1f, 1f);

            yield return null;
        }

        hitStartTime = Time.time;
        while ((Time.time - hitStartTime) <= GlobalSettings.Instance.CubicleObjectHitTime)
        {
            float zRotation = Mathf.SmoothStep(GlobalSettings.Instance.CubicleObjectHitRotation, 0f, (Time.time - hitStartTime) / GlobalSettings.Instance.CubicleObjectHitTime);
            float xScale = Mathf.SmoothStep(1f - GlobalSettings.Instance.CubicleObjectHitSquish, 1f, (Time.time - hitStartTime) / GlobalSettings.Instance.CubicleObjectHitTime);
			cubicleObject.transform.localEulerAngles = new Vector3(0f, 0f, zRotation * cubicleObjectHitDirection);
			if (spriteSwapper == null || !spriteSwapper.IsAnimating)
				cubicleObject.transform.localScale = new Vector3(xScale, 1f, 1f);

            yield return null;
        }

        // Make sure to set the final values as the lerps might not get all the way there
        ResetCubicleObject();

        yield return null;
    }

    private void ResetCubicleObject()
    {
        if (lastCubicleObject != null)
        {
            lastCubicleObject.transform.localEulerAngles = new Vector3(0f, 0f, 0f);
            if (spriteSwapper == null || !spriteSwapper.IsAnimating)
                lastCubicleObject.transform.localScale = new Vector3(1f, 1f, 1f);
        }
        
        spriteRenderer.sprite = IsMonkTime ? MonkNormalSprite : NormalSprite;
    }

    protected void StartTutorialShow(LevelConfig.Actions action, LevelConfig.CubiclePosition tutorialPosition)
    {
        if (tutorialCoroutine != null)
            StopCoroutine(tutorialCoroutine);
        
        tutorialCoroutine = StartCoroutine(TutorialShow(action, tutorialPosition));
    }

    protected IEnumerator TutorialShow(LevelConfig.Actions action, LevelConfig.CubiclePosition tutorialPosition)
    {
        if (action == LevelConfig.Actions.Up)
            TutorialIcon.sprite = GlobalSettings.Instance.SpriteTutorialUp;
        if (action == LevelConfig.Actions.LeftPosition)
            TutorialIcon.sprite = GlobalSettings.Instance.SpriteTutorialLeft;
        if (action == LevelConfig.Actions.RightPosition)
            TutorialIcon.sprite = GlobalSettings.Instance.SpriteTutorialRight;
        if (action == LevelConfig.Actions.CenterPosition && tutorialPosition == LevelConfig.CubiclePosition.Left)
            TutorialIcon.sprite = GlobalSettings.Instance.SpriteTutorialRight;
        if (action == LevelConfig.Actions.CenterPosition && tutorialPosition == LevelConfig.CubiclePosition.Right)
            TutorialIcon.sprite = GlobalSettings.Instance.SpriteTutorialLeft;

        if (tutorialPosition == LevelConfig.CubiclePosition.Left)
            TutorialGameObject.transform.localPosition = new Vector3(-GlobalSettings.Instance.MoveDistanceHorizontal, TutorialGameObject.transform.localPosition.y, TutorialGameObject.transform.localPosition.z);
        if (tutorialPosition == LevelConfig.CubiclePosition.Center)
            TutorialGameObject.transform.localPosition = new Vector3(0f, TutorialGameObject.transform.localPosition.y, TutorialGameObject.transform.localPosition.z);
        if (tutorialPosition == LevelConfig.CubiclePosition.Right)
            TutorialGameObject.transform.localPosition = new Vector3(GlobalSettings.Instance.MoveDistanceHorizontal, TutorialGameObject.transform.localPosition.y, TutorialGameObject.transform.localPosition.z);

        TutorialGameObject.SetActive(true);

        Vector3 startScale = Vector3.zero;
        Vector3 endScale = Vector3.one;

        float showStartTime = Time.time;
        while ((Time.time - showStartTime) <= GlobalSettings.Instance.TutorialShowTime)
        {
            TutorialGameObject.transform.localScale = Vector3.Lerp(startScale, endScale, Mathf.SmoothStep(0f, 1f, (Time.time - showStartTime) / GlobalSettings.Instance.TutorialShowTime));

            yield return null;
        }

        yield return new WaitForSeconds(GlobalSettings.Instance.TutorialVisibleTime - GlobalSettings.Instance.TutorialShowTime * 2f);

        showStartTime = Time.time;
        while ((Time.time - showStartTime) <= GlobalSettings.Instance.TutorialShowTime)
        {
            TutorialGameObject.transform.localScale = Vector3.Lerp(endScale, startScale, Mathf.SmoothStep(0f, 1f, (Time.time - showStartTime) / GlobalSettings.Instance.TutorialShowTime));

            yield return null;
        }

        ResetTutorial();

        yield return null;
    }

    private void ResetTutorial()
    {
        TutorialGameObject.SetActive(false);
    }
}
